@@ -1,5 +1,7 @@
 from __future__ import absolute_import
 
+import collections
+import opcode
 import os
 import threading
 from collections import defaultdict
@@ -13,7 +15,8 @@ from .util import CALL_COLORS
 from .util import CODE_COLORS
 from .util import MISSING
 from .util import OTHER_COLORS
-from .util import STRING_TYPES
+from .util import PY3
+from .util import StringType
 from .util import builtins
 from .util import iter_symbols
 from .util import safe_repr
@@ -23,10 +26,8 @@ try:
 except ImportError:
     from thread import get_ident
 
-
 __all__ = ['Action', 'Debugger', 'Manhole', 'CodePrinter', 'CallPrinter', 'VarsPrinter']
 
-
 BUILTIN_REPR_FUNCS = {
     'repr': repr,
     'safe_repr': safe_repr
@@ -38,11 +39,23 @@ class Action(object):
         raise NotImplementedError()
 
 
+class LazyImportPdb(object):
+    def __repr__(self):
+        return "<class 'pdb.Pdb'>"
+
+    __str__ = __repr__
+
+    def __call__(self, **kwargs):
+        from pdb import Pdb
+        return Pdb(**kwargs)
+
+
 class Debugger(Action):
     """
     An action that starts ``pdb``.
     """
-    def __init__(self, klass=config.Default('klass', lambda **kwargs: __import__('pdb').Pdb(**kwargs)), **kwargs):
+
+    def __init__(self, klass=config.Default('klass', LazyImportPdb()), **kwargs):
         self.klass = config.resolve(klass)
         self.kwargs = kwargs
 
@@ -110,7 +123,7 @@ class ColorStreamAction(Action):
         self.force_pid = config.resolve(force_pid)
         stream = config.resolve(stream)
         if stream is None:
-            stream = config.DEFAULT_STREAM
+            from . import _default_stream as stream
         self.stream = stream
         self.filename_alignment = config.resolve(filename_alignment)
         self.thread_alignment = config.resolve(thread_alignment)
@@ -150,7 +163,7 @@ class ColorStreamAction(Action):
 
     @stream.setter
     def stream(self, value):
-        if isinstance(value, STRING_TYPES):
+        if isinstance(value, StringType):
             if value in self._stream_cache:
                 value = self._stream_cache[value]
             else:
@@ -181,6 +194,27 @@ class ColorStreamAction(Action):
         else:
             raise TypeError('Expected a callable or either "repr" or "safe_repr" strings, not {!r}.'.format(value))
 
+    def try_str(self, obj):
+        """
+        Safely call ``str(obj)``. Failures will have special colored output and output is trimmed according
+        to ``self.repr_limit``.
+
+        Only used when dumping detached events.
+
+        Returns: string
+        """
+        limit = self.repr_limit
+        try:
+            s = str(obj)
+            s = s.replace('\n', r'\n')
+            if len(s) > limit:
+                cutoff = limit // 2
+                return '{} {CONT}[...]{RESET} {}'.format(s[:cutoff], s[-cutoff:], **self.other_colors)
+            else:
+                return s
+        except Exception as exc:
+            return '{INTERNAL-FAILURE}!!! FAILED REPR: {INTERNAL-DETAIL}{!r}{RESET}'.format(exc, **self.other_colors)
+
     def try_repr(self, obj):
         """
         Safely call ``self.repr_func(obj)``. Failures will have special colored output and output is trimmed according
@@ -362,7 +396,7 @@ class CodePrinter(ColorStreamAction):
                 self.filename_prefix(),
                 '...',
                 event.kind,
-                event.arg if event.detached else self.try_repr(event.arg),
+                self.try_str(event.arg) if event.detached else self.try_repr(event.arg),
                 COLOR=self.event_colors.get(event.kind),
             )
 
@@ -415,7 +449,7 @@ class CallPrinter(CodePrinter):
                 event.function,
                 ', '.join('{VARS}{VARS-NAME}{0}{VARS}={RESET}{1}'.format(
                     var,
-                    event.locals.get(var, MISSING) if event.detached else self.try_repr(event.locals.get(var, MISSING)),
+                    self.try_str(event.locals.get(var, MISSING)) if event.detached else self.try_repr(event.locals.get(var, MISSING)),
                     **self.other_colors
                 ) for var in code.co_varnames[:code.co_argcount]),
                 COLOR=self.event_colors.get(event.kind),
@@ -429,7 +463,7 @@ class CallPrinter(CodePrinter):
                 event.kind,
                 '   ' * (len(stack) - 1),
                 event.function,
-                event.arg if event.detached else self.try_repr(event.arg),
+                self.try_str(event.arg) if event.detached else self.try_repr(event.arg),
                 COLOR=self.event_colors.get(event.kind),
             )
 
@@ -442,7 +476,7 @@ class CallPrinter(CodePrinter):
                 event.kind,
                 '   ' * (len(stack) - 1),
                 event.function,
-                event.arg if event.detached else self.try_repr(event.arg),
+                self.try_str(event.arg) if event.detached else self.try_repr(event.arg),
                 COLOR=self.event_colors.get(event.kind),
             )
             if stack and stack[-1] == ident:
@@ -506,7 +540,7 @@ class VarsPrinter(ColorStreamAction):
             except Exception as exc:
                 printout = '{INTERNAL-FAILURE}FAILED EVAL: {INTERNAL-DETAIL}{!r}'.format(exc, **self.other_colors)
             else:
-                printout = obj if event.detached else self.try_repr(obj)
+                printout = self.try_str(obj) if event.detached else self.try_repr(obj)
 
             if frame_symbols >= symbols:
                 if first:
@@ -535,12 +569,21 @@ class VarsSnooper(ColorStreamAction):
     """
     A PySnooper-inspired action, similar to :class:`~hunter.actions.VarsPrinter`, but only show variable changes.
 
-    .. warning: Should be considered experimental. Use judiciously.
+    .. warning:: Should be considered experimental. Use judiciously.
 
         * It stores reprs for all seen variables, therefore it can use lots of memory.
         * Will leak memory if you filter the return events (eg: ``~Q(kind="return")``).
         * Not thoroughly tested. May misbehave on code with closures/nonlocal variables.
+
+    Args:
+        stream (file-like): Stream to write to. Default: ``sys.stderr``.
+        filename_alignment (int): Default size for the filename column (files are right-aligned). Default: ``40``.
+        force_colors (bool): Force coloring. Default: ``False``.
+        repr_limit (bool): Limit length of ``repr()`` output. Default: ``512``.
+        repr_func (string or callable): Function to use instead of ``repr``.
+            If string must be one of 'repr' or 'safe_repr'. Default: ``'safe_repr'``.
     """
+
     def __init__(self, **options):
         super(VarsSnooper, self).__init__(**options)
         self.stored_reprs = defaultdict(dict)
@@ -557,7 +600,7 @@ class VarsSnooper(ColorStreamAction):
         empty_filename_prefix = self.filename_prefix()
 
         current_reprs = {
-            name: value if event.detached else self.try_repr(value)
+            name: self.try_str(value) if event.detached else self.try_repr(value)
             for name, value in event.locals.items()
         }
         scope_key = event.code or event.function
@@ -614,3 +657,96 @@ class VarsSnooper(ColorStreamAction):
                     )
         if event.kind == 'return':
             del self.stored_reprs[scope_key]
+
+
+RETURN_VALUE = opcode.opmap['RETURN_VALUE']
+
+
+class ErrorSnooper(CodePrinter):
+    """
+    An action that prints events around silenced exceptions. Note that it inherits the output of :class:`~hunter.actions.CodePrinter` so no
+    fancy call indentation.
+
+    .. warning::
+
+        Should be considered experimental. May show lots of false positives especially if you're tracing lots of clumsy code like:
+
+        .. sourcecode:: python
+
+            try:
+                stuff = something[key]
+            except KeyError:
+                stuff = "default"
+
+    Args:
+        max_backlog (int): Maximum number of events to record and display before the silenced exception is raised.
+            Set to 0 to disable and get a speed boost. Default: ``10``.
+        max_events (int): Maximum number of events to record and display for each detected silenced exception. Default: ``50``.
+        max_depth (int): Increase if you want to drill into subsequent calls after an exception is raised. If you increase this you might
+            want to also increase ``max_events`` since subsequent calls may have so many events you won't get to see the return event.
+            Default: ``0`` (doesn't drill into any calls).
+
+        stream (file-like): Stream to write to. Default: ``sys.stderr``.
+        filename_alignment (int): Default size for the filename column (files are right-aligned). Default: ``40``.
+        force_colors (bool): Force coloring. Default: ``False``.
+        repr_limit (bool): Limit length of ``repr()`` output. Default: ``512``.
+        repr_func (string or callable): Function to use instead of ``repr``.
+            If string must be one of 'repr' or 'safe_repr'. Default: ``'safe_repr'``.
+
+    .. versionadded:: 3.1.0
+    """
+
+    def __init__(self, *args, **kwargs):
+        self.max_backlog = kwargs.pop('max_backlog', 10)
+        self.backlog = collections.deque(maxlen=self.max_backlog)
+        self.max_events = kwargs.pop('max_events', 50)
+        self.max_depth = kwargs.pop('max_depth', 0)
+        self.origin = None
+        self.events = None
+        super(ErrorSnooper, self).__init__(*args, **kwargs)
+
+    def __call__(self, event):
+        if self.max_backlog:
+            detached_event = event.detach(self.try_repr)
+            self.backlog.append(detached_event)
+        else:
+            detached_event = None
+
+        if event.kind == 'exception':  # something interesting happened ;)
+            if detached_event is None:
+                detached_event = event.detach(self.try_repr)
+            if self.origin:
+                self.events.append(detached_event)
+                if self.origin.depth > event.depth:
+                    self.origin = detached_event
+            else:
+                self.events = list(self.backlog)
+                self.origin = detached_event
+        elif self.origin:
+            if event.kind == 'return':
+                if detached_event is None:
+                    detached_event = event.detach(self.try_repr)
+                self.events.append(detached_event)
+                if event.depth == self.origin.depth - 1:  # stop if the same function returned (depth is -1)
+                    if (event.code.co_code[event.frame.f_lasti] if PY3 else ord(event.code.co_code[event.frame.f_lasti])) == RETURN_VALUE:
+                        self.dump_events()
+                        self.output("{BRIGHT}{fore(BLACK)}{} function exit{RESET}\n", "-" * 46)
+                    self.origin = None
+                    self.events = None
+            elif event.depth > self.origin.depth + self.max_depth:  # too many details
+                return
+            elif len(self.events) > self.max_events:
+                self.dump_events()
+                self.output("{BRIGHT}{fore(BLACK)}{} too many lines{RESET}\n", "-" * 46)
+            else:
+                if detached_event is None:
+                    detached_event = event.detach(self.try_repr)
+                self.events.append(detached_event)
+
+    def dump_events(self):
+        self.output("{BRIGHT}{fore(BLUE)}{} tracing {fore(YELLOW)}{}{fore(BLUE)} on {fore(RED)}{}{RESET}\n",
+                    ">" * 46, self.origin.function, self.origin.arg)
+        for event in self.events:
+            super(ErrorSnooper, self).__call__(event)
+        self.origin = None
+        self.events = None
